-module(dets_tests).

-include_lib("eqc/include/eqc.hrl").
-include_lib("eqc/include/eqc_statem.hrl").
-include("pulse.hrl").

-compile(export_all).

-define(DETS, dets).

-record(state, {objects, type, continuation}).		

%%Here the continuation will be a list of tuples which look like: {Continuation, MustReturnList, Pattern, Limit, MayReturnList}.

%%-------------------------------------------------------------------------------------------
%% Generators
%%-------------------------------------------------------------------------------------------

gen_ttype() ->
	oneof([set, bag, duplicate_bag]).

gen_object(S) ->
	frequency([
	{1,?LET(ListOfObjects, ?SUCHTHAT(L,
			list(?LET(Tuple,oneof([{char()}|S#state.objects]),oneof(tuple_to_list(Tuple)))), 
			L/=[]), 
		list_to_tuple(ListOfObjects))},
		{6, oneof([{char()}|S#state.objects])},
		{1, ?LET(KeyObject, gen_key(S), 
				?LET(Rest, list(char()), list_to_tuple([KeyObject|Rest])))}]).

gen_key(S) ->  
	?LET(Object, oneof([{char()}|S#state.objects]), element(1, Object)).	

gen_pattern(S) ->
	?LET(Pattern,
		?LET(Length,
			?LET(Object, gen_object(S), length(tuple_to_list(Object))), vector(Length, '_' )),
				list_to_tuple(Pattern)).

gen_limit() ->
	?SUCHTHAT(Limit, int(), Limit>0).

%%-------------------------------------------------------------------------------------------
%%-------------------------------------------------------------------------------------------
%%
%%	Test function:	
%%		dets:insert(Tab, Objects) -> ok | {error, Reason}
%%				Insert one or more objects into a DETS table
%%		dets:insert_new(Tab, Objects) -> Bool
%%				Insert one or more objects into a DETS table
%%		dets:lookup(Tab, Key) -> (Object) | {error, Reason}			
%%				Return all object with a given key in a DETS table
%%		dets:delete(Tab, Key) -> ok | {error, Reason}					
%%				Delete all objects with a given key from a DETS table
%%		dets:delete_all_objects(Tab) -> ok | {error, Reason}			
%%				Delete all objects from a DETS table
%%		dets:delete_object(Tab, Object)->ok | {error, Reason}			
%%				Delete a specific object in a DETS table
%%		dets:match_delete(Tab, Pattern) -> N | {error, Reason}
%%				Delete all objects which match a given pattern from a DETS table
%%		dets:match_object(Tab, Pattern) -> {Object} | {error, Reason}
%%				Match  the objects stored in a DETS table and return a list of objects
%%		dets:match_object(Tab, Pattern, Limit) -> {{Object}, Continuation}|'$end_of_table' 
%%							| {error, Reason}
%%				Match the first chunk of objects stored in a DETS table, and return a list 
%%				of objects
%%		dets:match_object(Continuation) -> {{Match}, Continuation|'$end_of_table'.
%%				Match a chunk of objects stored in a DETS table and return a list of objects
%%
%%-------------------------------------------------------------------------------------------
%%-------------------------------------------------------------------------------------------
%%	precondition
%%-------------------------------------------------------------------------------------------
%%-------------------------------------------------------------------------------------------

precondition(_S, {call, ?DETS, insert, [_TableId, _ObjectOrObjects]}) ->
	true;
precondition(_S, {call, ?DETS, insert_new, [_TableId, _ObjectOrObjects]}) ->
	true;
precondition(_S, {call, ?DETS, lookup, [_TableId, _Key]})->
	true;
precondition(_S, {call, ?DETS, delete, [_TableId, _Key]}) ->
	true;
precondition(_S, {call, ?DETS, delete_all_objects, [_TableId]}) ->
	true;
precondition(_S, {call, ?DETS, delete_object, [_TableId, _ObjectOrObjects]}) ->
	true;
precondition(_S, {call, ?DETS, match_delete, [_TableId, _Pattern]}) ->
	true;
precondition(_S, {call, ?DETS, match_object, [_TableId, _Pattern]}) ->
	true;
precondition(_S, {call, ?DETS, match_object, [_TableId, _Pattern, _Limit]}) ->
	true;
precondition(_S, {call, ?MODULE, my_match_object, [_Continuation]}) ->		
	true;
precondition(_S, {call,_,_,_}) ->
	true.

%%-------------------------------------------------------------------------------------------
%%-------------------------------------------------------------------------------------------
%%	next_state
%%-------------------------------------------------------------------------------------------
%%-------------------------------------------------------------------------------------------

next_state(S, _, {call, ?DETS, insert, [_TableId, [H|T]]}) ->	
	S1=next_state(S, 1, {call, ?DETS, insert, [_TableId, H]}),
	next_state(S1, 1, {call, ?DETS, insert, [_TableId, T]});
next_state(S, _, {call, ?DETS, insert, [_TableId, []]}) ->	
	S;
next_state(S, _, {call, ?DETS, insert, [_TableId, Object]}) ->		
	case S#state.type of				
		set ->
			Key = element(1, Object),
			S#state{objects=[Object | lists:keydelete(Key, 1, S#state.objects)], 
					continuation=insert_cont(Object, 
									delete_key_cont(Key, S#state.continuation))};
		bag ->
			case lists:member(Object, S#state.objects) of
				true -> S;
				false -> S#state{objects=[Object|S#state.objects],
									continuation=insert_cont(Object, S#state.continuation)}
			end;
		duplicate_bag ->
			S#state{objects=[Object|S#state.objects],
					continuation=insert_cont(Object, S#state.continuation)}
	end;
	
	
next_state(S, _, {call, ?DETS, insert_new, [_TableId, Object]}) ->		
	Key = element(1, Object),
	case my_lookup(Key, S#state.objects) of
		[] -> S#state{objects=[Object | S#state.objects], 
					continuation=insert_cont(Object, S#state.continuation)};
		_ -> S
	end;
	
	
next_state(S, _ObjectOrObjects, {call, ?DETS, lookup, [_TableId, _Key]}) ->	
	S;
next_state(S, _, {call, ?DETS, delete, [_TableId, Key]}) ->
	case lists:keymember(Key, 1, S#state.objects) of
		true -> S#state{objects = S#state.objects--my_lookup(Key, S#state.objects),
						continuation=delete_key_cont(Key, S#state.continuation)};
		false -> S
	end;
next_state(S, _, {call, ?DETS, delete_all_objects, [_TableId]}) ->		
	S#state{objects = [], 
			continuation = []};
next_state(S, _, {call, ?DETS, delete_object, [_TableId, Object]}) ->	
	S#state{objects = my_deleteobject(Object, S#state.objects),
			continuation = delete_object_cont(Object, S#state.continuation)};
next_state(S, _, {call, ?DETS, match_delete, [_TableId, Pattern]}) ->	
	S#state{objects = S#state.objects -- list_match_object(Pattern, S#state.objects),
			continuation = delete_pattern_cont(Pattern, S#state.continuation)};
next_state(S, _, {call, ?DETS, match_object, [_TableId, _Pattern]}) ->		
	S;
next_state(S, '$end_of_table', {call, ?DETS, match_object, [_TableId, _Pattern, _Limit]}) ->
	S;
next_state(S, {var, N}, {call, ?DETS, match_object, [_TableId, Pattern, Limit]}) ->	
	S#state{continuation = [{{{call, ?MODULE, my_element,  [{var, N}]}, []}, 
							 [], Pattern, Limit, []}|S#state.continuation]};
next_state(S, {Return, Cont}, {call, ?DETS, match_object, [_TableId, Pattern, Limit]}) ->	
	case Cont of
		'$end_of_table' -> S;
		_ -> RemainReturnList=list_match_object(Pattern, S#state.objects) -- Return, 
			 S#state{continuation = 
						[{{Cont, RemainReturnList},  RemainReturnList, 
							Pattern, Limit, []} | S#state.continuation]}
	end;
next_state(S, '$end_of_table', {call, ?MODULE, my_match_object, [_PreCont]}) ->
	S;
next_state(S, {var, N}, {call, ?MODULE, my_match_object, [_PreCont]}) ->	
	S#state{continuation = [{{{call, ?MODULE, my_element,  [{var, N}]}, []}, 
							[], {}, undefined, []}|S#state.continuation]};
next_state(S, {Return, Cont}, {call, ?MODULE, my_match_object, [PreCont]}) ->
	{{PreCont, RemainMatchList}, MustReturnList, Pattern, Limit, MayReturnList}
		= my_keyfind(PreCont, S#state.continuation),
	case Cont of
		'$end_of_table'-> S;
		_ -> S#state{continuation = 
				[{{Cont, RemainMatchList--Return}, MustReturnList -- Return, Pattern, Limit, 
					MayReturnList--Return} | S#state.continuation]}
	end;
next_state(S, _R, {call, ?DETS, open_file, [test, [{type, _Type}]]})->
	S.

%%-------------------------------------------------------------------------------------------
%%-------------------------------------------------------------------------------------------
%%	postcondition.
%%-------------------------------------------------------------------------------------------
%%-------------------------------------------------------------------------------------------

postcondition(_S, {call, ?DETS, insert, [_TableId, _ObjectOrObjects]}, R) ->
	R == ok;
postcondition(S, {call, ?DETS, insert_new, [_TableId, Object]}, R) ->
	Key = element(1, Object),
	case my_lookup(Key, S#state.objects) of
		[] -> R==true;
		_ -> R==false
	end;
postcondition(S, {call, ?DETS, lookup, [_TableId, Key]}, R) ->
	lists:sort(R) == lists:sort(my_lookup(Key, S#state.objects));
postcondition(_S, {call, ?DETS, delete, [_TableId, _Key]}, R) ->
	R == ok;
postcondition(_S, {call, ?DETS, delete_all_objects, [_TableId]}, R) ->
	R == ok;
postcondition(_S, {call, ?DETS, delete_object, [_TableId, _ObjectOrObjects]}, R) ->
	R == ok;
postcondition(_S, {call, ?DETS, match_delete, [_TableId, _Pattern]}, R) ->
	R == ok;
postcondition(S, {call, ?DETS, match_object, [_TableId, Pattern]}, R) ->
	lists:sort(R) == lists:sort(list_match_object(Pattern, S#state.objects));
postcondition(S, {call, ?DETS, match_object, [_TableId, Pattern, _Limit]}, 
					'$end_of_table') ->
	list_match_object(Pattern, S#state.objects) == []; 
postcondition(S, {call, ?DETS, match_object, [_TableId, Pattern, _Limit]}, {MatchList, _}) ->
			check_inclusion(MatchList, list_match_object(Pattern, S#state.objects));
postcondition(S, {call, ?MODULE, my_match_object, [PreCont]}, '$end_of_table') ->
%	io:format("~p~n~p~n", [PreCont, S#state.continuation]),
	case PreCont of
		'$end_of_table'-> true;
		_ -> 
%		io:format("~p~n", [my_keyfind(PreCont, S#state.continuation)]), 
		element(2, my_keyfind(PreCont, S#state.continuation))==[]
	end;
postcondition(S, {call, ?MODULE, my_match_object, [PreCont]}, {MatchList, Cont}) ->
	{{PreCont, _RemainMatchList}, MustReturnList, _Pattern, _Limit, MayReturnList}
		=my_keyfind(PreCont, S#state.continuation),
	case Cont of
		'$end_of_table' -> check_inclusion(MustReturnList, MatchList) andalso
							check_inclusion(MatchList, MustReturnList++MayReturnList);
		_ -> check_inclusion(MatchList, MustReturnList++MayReturnList)
	end;
postcondition(_S, {call, ?DETS, open_file, [test, [{type, _Type}]]}, _R)->
	true.
				
%%-------------------------------------------------------------------------------------------
%%-------------------------------------------------------------------------------------------
%%	command
%%-------------------------------------------------------------------------------------------
%%-------------------------------------------------------------------------------------------

command(S) ->
	frequency(
		  [{5, {call, ?DETS, insert, [{var, tableId}, 
			oneof([list(gen_object(S)), gen_object(S)])]}}]++
		  [{5, {call, ?DETS, insert_new, [{var, tableId}, 
			gen_object(S)]}}]++
		  [{1, {call, ?DETS, lookup, [{var, tableId}, gen_key(S)]}}]++
		  [{2, {call, ?DETS, delete, [{var, tableId}, gen_key(S)]}}]++
		  [{2, {call, ?DETS, delete_all_objects, [{var, tableId}]}}]++
		  [{3, {call, ?DETS, delete_object, [{var, tableId}, gen_object(S)]}}]++
		  [{2, {call, ?DETS, match_delete, [{var, tableId}, gen_pattern(S)]}}]++
		  [{5, {call, ?DETS, match_object, [{var, tableId}, gen_pattern(S)]}}]++
		  [{5, {call, ?DETS, match_object, [{var, tableId}, gen_pattern(S), gen_limit()]}}]++
		  [{5, ?LET(Continuation,  oneof(S#state.continuation),{call, ?MODULE, my_match_object,
				[{call, erlang, element, 
					[1, {call, erlang, element,[1, Continuation]}]}]})}|| S#state.continuation/=[]]
			).

%%-------------------------------------------------------------------------------------------
%%-------------------------------------------------------------------------------------------
%%	Property
%%-------------------------------------------------------------------------------------------
%%-------------------------------------------------------------------------------------------

delete_file(File) ->
	case file:delete(File) of
		ok -> delete_file(File);
		{error, enoent} -> ok
	end.
	
bug_commands() ->
	{[{init,#state{objects=[],type=bag,continuation=[]}}],
	 {[],
	  [{set, {var, 2}, {call, ?DETS, insert, [{var, tableId}, {0,0}]}},
	   {set, {var, 3}, {call, ?DETS, match_object, [{var, tableId}, {'_', '_'}]}}]}}.
	
prop_dets_test() ->
	?FORALL(Type,  gen_ttype(),
	
%%-----PROPERTY FOR SEQENTIAL TESTS--------	
		?FORALL(Cmds, commands(?MODULE, #state{objects = [], type = Type, continuation=[]}),
				begin
					dets_eqc_cleanup:cleanup(),
					?DETS:close(test),
					delete_file("test"),
					
					{ok, TableId}=?DETS:open_file(test, [{type, Type}]),
%					?DETS:safe_fixtable(TableId, true),
					{H, S, Res}=run_commands(?MODULE, Cmds, [{tableId, TableId}]),
					?WHENFAIL(io:format("~p\n~p\n~p\n", [H, S, Res]),
										Res==ok)
				end)).
		
		
%%-----PROPERTY FOR PULSE TESTS--------		
%		begin
%			?ALWAYS(50,
%			?FORALL(Cmds, %bug_commands(), %
%			parallel2:pcommands(?MODULE, #state{objects = [], type = Type, continuation=[]}),
%				?SCHEDULED([{verbose, false}], 
%					begin
%						%{L1,L2}=Cmds,
%						%L3=[[{set, {var, 0},{call, ?DETS, open_file, [test, [{type, 
%						%		Type}]]}}|X] || X <-L2],
%						%Cmds2={L1, L3},
%						{L1,{L2,L3}}=Cmds,
%						Cmds2={L1,
%							{[{set, {var, 0},{call, ?DETS, open_file, 
%														[test, [{type, Type}]]}}|L2],
%							[{set, {var, 0},{call, ?DETS, open_file, 
%														[test, [{type, Type}]]}}|L3]}},
%						dets_eqc_cleanup:cleanup(),
%						?DETS:close(test),
%						delete_file("test"),
%						d_ets:start(),
%						{ok, TableId}=?DETS:open_file(test, [{type, Type}]),
%				%		?DETS:safe_fixtable(TableId, true),
%						parallel2:run_pcommands(?MODULE, Cmds2, [{tableId, TableId}])
%					end,
%					X,
%					begin
%	%					?WHENFAIL(io:format("~p~n", [X]),
%	%					begin
%							kill_all_pids:kill_all_pids(X),
%							{H,S,Res} = protest_pulse:result(X),
%							?WHENFAIL(io:format("~p\n~p\n~p\n", [H, S, Res]),
%										Res==ok)
%	%					end)
%					end)))		
%		end).

%%-------------------------------------------------------------------------------------------
%%-------------------------------------------------------------------------------------------
%% Help function
%%
%%	my_lookup(Key, List) 
%%				-> Find all the elements with the same Key in the List.
%%	list_match_object(Pattern, List) 
%%				-> Return a list of elements match with the Pattern.
%%	check_inclusion(L, List) 
%%				-> Check if the members in L are still in List.
%%	my_element('$end_of_table' | {_, Cont})
%%				->When the parameter is an atom, it return an atom, when is a tuple with two elements, it will return the second element.
%%	my_deleteobject(Object, List)	
%%				->Recursively delete an object from List.
%%
%%----THE LIST [H | T] BELOW REFERS TO THE CONTINUATION IN THE STATE, WHICH WITH FOUR ELEMENTS.----%%
%%
%%	delete_key_cont(Key, [H|T]) 
%%				->Delete the elements with given Key from RemainMatchList in the continuation in state.
%%	delete_object_cont(Object,[H|T])
%%				->Delete one object from RemainMatchList in the continuation in state.
%%	insert_cont(Object,[H|T])
%%				->Insert an object into the RemainMatchList in the continuation in state.
%%	delete_pattern_cont(Pattern, [H|T])
%%				->Delete the Continuations with Pattern, from the continuation in state.
%%
%%-------------------------------------------------------------------------------------------
%%-------------------------------------------------------------------------------------------

my_lookup(Key, [H|T]) ->
	case Key==element(1,H) of
		true  -> [H | my_lookup(Key, T)];
		false -> my_lookup(Key, T)
	end;
my_lookup(_Key, []) ->
	[].

list_match_object(Pattern, [H|T]) ->
	case length(tuple_to_list(H)) == length(tuple_to_list(Pattern)) of
		true -> [H|list_match_object(Pattern, T)];
		false -> list_match_object(Pattern, T)
	end;
list_match_object(_Pattern, []) ->
	[].
	
check_inclusion([H|T], List) ->
	case lists:member(H, List) of
		true -> check_inclusion(T, List);
		false -> false
	end;
check_inclusion([], _List)->
	true.
	
my_element('$end_of_table')->
	'$end_of_table';
my_element({_Reture, Cont}) ->
	Cont.
	
my_deleteobject(Object, [H|T])->
	case Object==H of 
		true -> my_deleteobject(Object, T);
		false-> [H | my_deleteobject(Object, T)]
	end;
my_deleteobject(_Object, [])->
	[].
	
my_keyfind(Key, [H|T])->
	case Key==element(1, element(1, H)) of
		true-> H;
		false-> my_keyfind(Key, T)
	end;
my_keyfind(_Key, [])->
	[].

same_find([H|T], List)->
	case lists:member(H, List) of
		true-> [H|same_find(T, List--[H])];
		false-> same_find(T, List)
	end;
same_find([], _List)->
	[].

%%------------------------------------------------------------------------	
%%-----HERE THE LIST [H|T] IS THE CONTINUATION LIST IN THE STATE.---------
%%------------------------------------------------------------------------

insert_cont(Object, [H|T])->
	case length(tuple_to_list(Object))==length(tuple_to_list(element(3, H))) of
		true -> [{element(1, H), element(2, H), element(3, H), element(4, H), 
					element(5, H)++[Object]} | insert_cont(Object, T)];
		false -> [H	| insert_cont(Object, T)]
	end;	
insert_cont(_Object, [])->
	[].	

delete_key_cont(Key, [H|T]) ->
	case my_lookup(Key, element(2, H)++element(5, H))==[] of
		true -> [H | delete_key_cont(Key, T)];
		false-> 
			case my_lookup(Key, element(2, H)) of 
				[]-> MayReturn =
						lists:append(element(5,H)--my_lookup(Key, element(5, H)), 	
									same_find(my_lookup(Key, element(5, H)), 
												element(2, element(1, H)))),
						[{element(1,H), element(2, H), element(3,H), element(4, H), MayReturn} | delete_key_cont(Key, T)];
				_-> MayReturn =
						lists:append(my_lookup(Key, element (2, H)),
							lists:append(element(5,H)--my_lookup(Key, element(5, H)), 	
									same_find(my_lookup(Key, element(5, H)), 
												element(2, element(1, H))))),
						[{element(1,H), element(2, H)--my_lookup(Key, element(2, H)), 
						  element(3,H), element(4, H), MayReturn}	
						| delete_key_cont(Key, T)]
			end
	end;
delete_key_cont(_Key, []) ->
	[].	
	
delete_object_cont(Object,[H|T])->
	case lists:member(Object, element(2, H)++element(5, H)) of
		true ->	
			case lists:member(Object, element(2, H)) of
				true -> [{element(1, H), element(2, H)--[Object], 
						  element(3, H), element(4, H), [Object|element(5, H)]} 
						| delete_object_cont(Object, T)];
				false -> 
					case lists:member(Object, element(2, element(1, H))) of
						true-> [H | delete_object_cont(Object, T)];
						false-> [{element(1, H), element(2, H), element(3, H), element(4, H),
								 [Object|element(5, H)]} 
								| delete_object_cont(Object, T)]
					end
			end;
			
		false -> [ H | delete_object_cont(Object, T)]
	end;	
delete_object_cont(_Object, [])->
	[].

delete_pattern_cont(Pattern, [H|T])->
	case Pattern == element(3, H) of 
		true -> [{element(1, H),[], element(3, H), element(4, H), 
					element(5, H)++element(2, H)} 
				|delete_pattern_cont(Pattern, T)];
		false -> [H|delete_pattern_cont(Pattern, T)]
	end;
delete_pattern_cont(_Pattern, []) ->
	[].

%%-------------------------------------------------------------------------------
%%-------------------------------------------------------------------------------

my_match_object(Cont)->
	case Cont of
		'$end_of_table' -> '$end_of_table';
		_->?DETS:match_object(Cont)
	end.

compile()->
	instrument:c(genserver),
	instrument:c(super_visor),
	instrument:c(d_ets),
	instrument:c(d_ets_server),
	instrument:c(d_ets_sup),
	instrument:c(d_ets_utils),
	instrument:c(d_ets_v9),
	instrument:c(parallel2).
