-module(ets_tests).

-include_lib("eqc/include/eqc.hrl").
-include_lib("eqc/include/eqc_statem.hrl").

-compile(export_all).

-record(state, {objects, type, continuation, matchcont}).		

%%-------------------------------------------------------------------------------------------
%%-------------------------------------------------------------------------------------------
%%
%%	objects: 
%%		Save entire objects in ETS table.
%%	type: 
%%		Save the type of ETS table generated during test.
%%	continuation: 
%%		A list of tuples with structure: {{Continuation, RemainMatchList}, 
%%		MustReturnList, Pattern, Limit, MayReturnList}.
%%	matchcont: 
%%		A list of tuples with structure: {{MCont, ReturnList}, MustReturnList, Pattern, 
%%		Limit, MayReturnList}.
%%
%%-------------------------------------------------------------------------------------------
%%-------------------------------------------------------------------------------------------

%%-------------------------------------------------------------------------------------------
%%	Generators
%%-------------------------------------------------------------------------------------------

gen_ttype() ->
	oneof([set, ordered_set, bag, duplicate_bag]).

gen_object(S) ->
	frequency([
	{1,?LET(ListOfObjects, ?SUCHTHAT(L,
			list(?LET(Tuple,oneof([{char()}|S#state.objects]),oneof(tuple_to_list(Tuple)))), 
			L/=[]), 
		list_to_tuple(ListOfObjects))},
	{6, oneof([{char()}|S#state.objects])},
	{1, ?LET(KeyObject, gen_key(S), 
				?LET(Rest, list(char()), list_to_tuple([KeyObject|Rest])))}]).

gen_key(S) -> 
	?LET(Object, oneof([{char()}|S#state.objects]), element(1, Object)).	

gen_pattern(S) ->
	?LET(Pattern,
		 ?LET(Length,
			  ?LET(Object, gen_object(S), length(tuple_to_list(Object))), 
			  vector(Length, '_' )),
		 list_to_tuple(Pattern)).

gen_limit() ->
	?SUCHTHAT(Limit, int(), Limit>0).
	
gen_mcont(S) ->
	?LET(MatchCont, 
		 oneof([{{list_to_atom(io_lib:write(make_ref())), 			
				[]},[],[],undefined,[]}|S#state.matchcont]), 	
		 element(1, element(1,MatchCont))).
	
%%-------------------------------------------------------------------------------------------
%%-------------------------------------------------------------------------------------------
%%
%%	Test function:	
%%		ets:insert(Tab, ObjectOrObjects) -> ture.
%%				Insert an object into an ETS table
%%		ets:lookup(Tab, Key) -> (Object).
%%				Return all object with a given key in an ETS table
%%		ets:delete(Tab, Key) -> true.
%%				Delete all objects with a given key from an ETS table
%%		ets:delete_all_objects(Tab) -> true.			
%%				Delete all objects in an ETS table
%%		ets:delete_object(Tab, Object)->true.			
%%				Delete a specific object in an ETS table
%%		ets:match_delete(Tab, Pattern) -> true.
%%				Delete all objects which match a given pattern from an ETS table
%%		ets:match_object(Continuation) -> {{Match}, Continuation|'$end_of_table'}.
%%				Continuation matching objects in an ETS table)
%%		ets:match_object(Tab, Pattern) -> {Object}.
%%				Match the objects in an ETS table aganinst a pattern
%%		ets:match_object(Tab, Pattern, Limit) -> {{Match}, Continuation|'$end_of_table'}.
%%				Match the objects in an ETS table against a pattern, and return part of 
%%				the answer.
%%		ets:tab2list(Tab)->{Object}
%%				Return a lis of all objects in an ETS table.
%%				
%%-------------------------------------------------------------------------------------------
%%-------------------------------------------------------------------------------------------
%%	precondition.
%%-------------------------------------------------------------------------------------------

precondition(_S, {call, ets, insert, [_TableId, _ObjectOrObjects]}) ->
	true;
precondition(_S, {call, ets, lookup, [_TableId, _Key]}) ->
	true;
precondition(_S, {call, ets, delete, [_TableId, _Key]}) ->
	true;
precondition(_S, {call, ets, delete_all_objects, [_TableId]}) ->
	true;
precondition(_S, {call, ets, delete_object, [_TableId, _Object]}) ->
	true;
precondition(_S, {call, ets, match_delete, [_TableId, _Pattern]}) ->
	true;
precondition(S, {call, ?MODULE, start_match_object, [_TableId, _Pattern]}) ->
	S#state.matchcont==[];
precondition(S, {call, ?MODULE, finish_match_object, [_MCont]}) ->
	S#state.matchcont/=[];
precondition(_S, {call, ets, match_object, [_TableId, _Pattern, _Limit]}) ->
	true;
precondition(_S, {call, ets, match_object, [_Continuation]}) ->		
	true;
precondition(_S, {call, ?MODULE, my_tab2list, [_TableId]}) ->
	true;
precondition(_S, {call,_,_,_}) ->
	true.

%%-------------------------------------------------------------------------------------------
%%	next_state
%%-------------------------------------------------------------------------------------------
	
next_state(S, _, {call, ets, insert, [_TableId, [H|T]]}) ->	
	S1=next_state(S, 1, {call, ets, insert, [_TableId, H]}),
	next_state(S1, 1, {call, ets, insert, [_TableId, T]});
next_state(S, _, {call, ets, insert, [_TableId, []]}) ->	
	S;
next_state(S, _, {call, ets, insert, [_TableId, Object]}) ->		
	case S#state.type of
		set ->
			Key = element(1, Object),
			S#state{objects=[Object | lists:keydelete(Key, 1, S#state.objects)], 
					continuation=insert_cont(Object, 
									delete_key_cont(Key, S#state.continuation)),
					matchcont=insert_cont(Object, 
									delete_key_mcont(Key, S#state.matchcont))};
		ordered_set ->
			Key = element(1,Object),
			S#state{objects=[Object | lists:keydelete(Key, 1, S#state.objects)],
					continuation= insert_cont(Object, 
									delete_key_cont(Key, S#state.continuation)),
					matchcont=insert_cont(Object, 
									delete_key_mcont(Key, S#state.matchcont))};	
		bag ->
			case lists:member(Object, S#state.objects) of
				true -> S;
				false -> S#state{objects=[Object | S#state.objects],
									continuation=insert_cont(Object, S#state.continuation),
									matchcont=insert_cont(Object, S#state.matchcont)}
			end;
		duplicate_bag ->
			S#state{objects=[Object | S#state.objects],
					continuation=insert_cont(Object, S#state.continuation),
					matchcont=insert_cont(Object, S#state.matchcont)}
	end;
next_state(S, _ObjectOrObjects, {call, ets, lookup, [_TableId, _Key]}) ->	
	S;
next_state(S, _, {call, ets, delete, [_TableId, Key]}) ->
	case lists:keymember(Key, 1, S#state.objects) of
		true -> S#state{objects = S#state.objects--my_lookup(Key, S#state.objects),
						continuation=delete_key_cont(Key, S#state.continuation),
						matchcont=delete_key_mcont(Key, S#state.matchcont)};
		false -> S
	end;
next_state(S, _, {call, ets, delete_all_objects, [_TableId]}) ->	
	case S#state.matchcont of
		[{{MCont, ReturnList}, MustReturnList, Pattern, Some, MayReturnList}] ->
			S#state{objects = [], 
					continuation = [], 
					matchcont=[{{MCont, ReturnList}, [], Pattern, Some, 
								MustReturnList++MayReturnList}]};
		[] -> S#state{objects = [],
						continuation = []}
	end;
next_state(S, _, {call, ets, delete_object, [_TableId, Object]}) ->	
	S#state{objects = my_deleteobject(Object, S#state.objects),
			continuation = delete_object_cont(Object, S#state.continuation),
			matchcont = delete_object_mcont(Object, S#state.matchcont)};
next_state(S, _, {call, ets, match_delete, [_TableId, Pattern]}) ->	
	S#state{objects = S#state.objects -- list_match_object(Pattern, S#state.objects),
			continuation = delete_pattern_cont(Pattern, S#state.continuation),
			matchcont = delete_pattern_mcont(Pattern, S#state.matchcont)};
next_state(S, R, {call, ?MODULE, start_match_object, [_TableId, Pattern]}) ->		
	S#state{matchcont=[{R, list_match_object(Pattern, S#state.objects), 
						 Pattern, undefined, []}]};
next_state(S, _, {call, ?MODULE, finish_match_object, [_MCont]}) ->		
	S#state{matchcont=[]};
next_state(S, '$end_of_table', {call, ets, match_object, [_TableId, _Pattern, _Limit]}) ->
	S;
next_state(S, {var, N}, {call, ets, match_object, [_TableId, Pattern, Limit]}) ->	
	S#state{continuation = [{{{call, ?MODULE, my_element, [{var, N}]}, []}, 
							 [],Pattern, Limit, []} | S#state.continuation]};
next_state(S, {Return, Cont}, {call, ets, match_object, [_TableId, Pattern, Limit]}) ->	
	case Cont of
		'$end_of_table' -> S;
		_ -> RemainReturnList=list_match_object(Pattern, S#state.objects) -- Return, 
				S#state{continuation = 
						[{{Cont, RemainReturnList},  RemainReturnList, 
							Pattern, Limit, []} | S#state.continuation]}
	end;
next_state(S, '$end_of_table', {call, ets, match_object, [_PreCont]}) ->
	S;
next_state(S, {var, N}, {call, ets, match_object, [_PreCont]}) ->	
	S#state{continuation = [{{{call, ?MODULE, my_element,  [{var, N}]}, []}, 
							[], {}, undefined, []}|S#state.continuation]};
next_state(S, {Return, Cont}, {call, ets, match_object, [PreCont]}) ->
	{{PreCont, RemainMatchList}, MustReturnList, Pattern, Limit, MayReturnList}
		= my_keyfind(PreCont, S#state.continuation),
	case Cont of
		'$end_of_table'-> S;
		_ -> S#state{continuation = 
				[{{Cont, RemainMatchList--Return}, MustReturnList -- Return, Pattern, Limit, 
					MayReturnList--Return} | S#state.continuation]}
	end;
next_state(S, _, {call, ?MODULE, my_tab2list, [_TableId]})->
	S;
next_state(S, _, {call, ?MODULE, sleep, [_Time]})->
	S.
		
%%-------------------------------------------------------------------------------------------
%%	postcondition
%%-------------------------------------------------------------------------------------------

postcondition(_S, {call, ets, insert, [_TableId, _ObjectOrObjects]}, R) ->
	R == true;
postcondition(S, {call, ets, lookup, [_TableId, Key]}, R) ->
	lists:sort(R) == lists:sort(my_lookup(Key, S#state.objects));	
postcondition(_S, {call, ets, delete, [_TableId, _Key]}, R) ->
	R == true;
postcondition(_S, {call, ets, delete_all_objects, [_TableId]}, R) ->
	R == true;
postcondition(_S, {call, ets, delete_object, [_TableId, _ObjectOrObjects]}, R) ->
	R == true;
postcondition(_S, {call, ets, match_delete, [_TableId, _Pattern]}, R) ->
	R == true ;
postcondition(_S, {call, ?MODULE, start_match_object, [_TableId, _Pattern]}, _R) ->
	true;
postcondition(S, {call, ?MODULE, finish_match_object, [MCont]}, R) ->
	{MCont, MustReturnList, _Pattern, _, MayReturnList}
			=lists:keyfind(MCont, 1, S#state.matchcont),
	check_inclusion(R, MustReturnList++MayReturnList);
postcondition(S, {call, ets, match_object, [_TableId, Pattern, _Limit]}, '$end_of_table') ->
	list_match_object(Pattern, S#state.objects) == []; 
postcondition(S, {call, ets, match_object, [_TableId, Pattern, Limit]}, {MatchList, _}) ->
	length(MatchList) == erlang:min(Limit, length(list_match_object(Pattern, 
									S#state.objects))) andalso			
	check_inclusion(MatchList, list_match_object(Pattern, S#state.objects));
postcondition(S, {call, ets, match_object, [PreCont]}, '$end_of_table') ->
	case PreCont of
		'$end_of_table'-> true;
		_ -> element(2, my_keyfind(PreCont, S#state.continuation))==[]	
	end;
postcondition(S, {call, ets, match_object, [PreCont]}, {MatchList, Cont}) ->	
	{{PreCont, _RemainMatchList}, MustReturnList, _Pattern, Limit, MayReturnList}
		=my_keyfind(PreCont, S#state.continuation),
	case Cont of
		'$end_of_table' -> check_inclusion(MustReturnList, MatchList) andalso
							check_inclusion(MatchList, MustReturnList++MayReturnList);
		_ -> length(MatchList) =< Limit	andalso	
					check_inclusion(MatchList, MustReturnList++MayReturnList)
	end;
postcondition(S, {call,?MODULE, my_tab2list, [_TableId]}, R) ->
	lists:sort(R)==lists:sort(S#state.objects).

	
%%-------------------------------------------------------------------------------------------
%%	command
%%-------------------------------------------------------------------------------------------

command(S) ->
	case S#state.matchcont==[] of
		true->
			frequency(
		  [{5, {call, ets, insert, [{var, tableId}, 
			oneof([list(gen_object(S)), gen_object(S)])]}}]++				
		  [{1, {call, ets, lookup, [{var, tableId}, gen_key(S)]}}]++
		  [{2, {call, ets, delete, [{var, tableId}, gen_key(S)]}}]++
		  [{3, {call, ets, delete_all_objects, [{var, tableId}]}}]++
		  [{2, {call, ets, delete_object, [{var, tableId}, gen_object(S)]}}]++
		  [{2, {call, ets, match_delete, [{var, tableId}, gen_pattern(S)]}}]++
		  [{5, {call, ?MODULE, start_match_object, [{var, tableId}, gen_pattern(S)]}}]++  
		  [{5, {call, ets, match_object, [{var, tableId}, gen_pattern(S), gen_limit()]}}]++
		  [{3, ?LET(Continuation,  oneof(S#state.continuation),{call, ets, match_object, 
				[element(1,element(1,Continuation))]})}|| S#state.continuation/=[]]++
		  [{5, {call, ?MODULE, my_tab2list, [{var, tableId}]}}]
		  );
		false->
			?LET(MCont, oneof(S#state.matchcont), {call, ?MODULE, finish_match_object, [{call, erlang, element,[1, MCont]}]})
	end.
%%-------------------------------------------------------------------------------------------
%%	property
%%-------------------------------------------------------------------------------------------

prop_ets_test() ->
	?FORALL(Bool1, bool(), 
	?FORALL(Bool2, bool(), 
	?FORALL(Type, gen_ttype(), 
		begin
%%			?FORALL(Cmds, commands(?MODULE, #state{objects = [], type = Type, continuation=[], matchcont=[]}),
			?FORALL(Cmds, parallel_commands(?MODULE, #state{objects = [], type = Type, continuation=[], matchcont=[]}), 
				?ALWAYS(100, prop_ets_test(Bool1, Bool2, Type ,Cmds)))
		end))).
	
prop_ets_test(Bool1, Bool2, Type ,Cmds)->
		?LAZY(	begin
					TableId=ets:new(test, [public, Type, {write_concurrency, Bool1}, {read_concurrency, Bool2}]),
%%						{H,S,Res} = run_commands(?MODULE, Cmds, [{tableId, TableId}]),
						{H,S,Res} = run_parallel_commands(?MODULE, Cmds, [{tableId, TableId}]),
					ets:delete(TableId),
						?WHENFAIL(io:format("~p\n~p\n~p\n", [H, S, Res]),
							Res == ok)
				end).

prop_ets_test([H,T])->
	prop_ets_test(false, false, H, T).


%%-------------------------------------------------------------------------------------------
%%-------------------------------------------------------------------------------------------
%% Help function
%%
%%	my_lookup(Key, List) 
%%				-> Find all the elements with the same Key in the List.
%%	list_match_object(Pattern, List) 
%%				-> Return a list of elements match with the Pattern.
%%	check_inclusion(L, List) 
%%				-> Check if the members in L are still in List.
%%	my_element('$end_of_table' | {_, Cont})
%%				->When the parameter is an atom, it return an atom, when is a tuple with two elements, it will return the second element.
%%	my_deleteobject(Object, List)	
%%				->Recursively delete an object from List.
%%
%%----THE LIST [H | T] BELOW REFERS TO THE CONTINUATION IN THE STATE, WHICH WITH FIVE ELEMENTS.----%%
%%
%%	delete_key_cont(Key, [H|T]) 
%%				->Delete the elements with given Key from either RemainMatchList or MayReturnList in the continuation in state.
%%	delete_object_cont(Object,[H|T])
%%				->Delete one object from either RemainMatchList or MayReturnList in the continuation in state.
%%	insert_cont(Object,[H|T])
%%				->Insert an object into the MayReturnList in the continuation in state.
%%	delete_pattern_cont(Pattern, [H|T])
%%				->Delete the Continuations with Pattern, from the continuation in state.
%%
%%-------------------------------------------------------------------------------------------
%%-------------------------------------------------------------------------------------------

my_lookup(Key, [H|T]) ->
	case Key==element(1,H) of
		true  -> [H | my_lookup(Key, T)];
		false -> my_lookup(Key, T)
	end;
my_lookup(_Key, []) ->
	[].
	%%lists:keysearch/3 or lists:keyfind/3, can provide similar result, but just can provide the first matched tuple, which is not enough.
	
list_match_object(Pattern, [H|T]) ->
	case length(tuple_to_list(H)) == length(tuple_to_list(Pattern)) of
		true -> [H|list_match_object(Pattern, T)];
		false -> list_match_object(Pattern, T)
	end;
list_match_object(_Pattern, []) ->
	[].
	
check_inclusion([H|T], List) ->
	case lists:member(H, List) of
		true -> check_inclusion(T, List);
		false -> false
	end;
check_inclusion([], _List)->
	true.

my_element('$end_of_table')->
	'$end_of_table';
my_element({_Reture, Cont}) ->
	Cont.
	
my_deleteobject(Object, [H|T])->
	case Object==H of 
		true -> my_deleteobject(Object, T);
		false-> [H | my_deleteobject(Object, T)]
	end;
my_deleteobject(_Object, [])->
	[].
	
my_keyfind(Cont, [H|T])->
	case Cont==element(1, element(1, H)) of
		true-> H;
		false-> my_keyfind(Cont, T)
	end;
my_keyfind(_Cont, [])->
	[].

same_find([H|T], List)->
	case lists:member(H, List) of
		true-> [H|same_find(T, List--[H])];
		false-> same_find(T, List)
	end;
same_find([], _List)->
	[].
	
%%-------------------------------------------------------------------------------------------
%%-----HERE THE LIST [H|T] IS THE conttinuation LIST IN THE state.
%%-------------------------------------------------------------------------------------------

insert_cont(Object, [H|T])->
	case length(tuple_to_list(Object))==length(tuple_to_list(element(3, H))) of
		true -> [{element(1, H), element(2, H), element(3, H), element(4, H), 
				 [Object | element(5, H)]} 	| insert_cont(Object, T)];
		false -> [H	| insert_cont(Object, T)]
	end;	
insert_cont(_Object, [])->
	[].	

delete_key_cont(Key, [H|T]) ->
	case my_lookup(Key, element(2, H)++element(5, H))==[] of
		true -> [H | delete_key_cont(Key, T)];
		false-> 
			case my_lookup(Key, element(2, H)) of 
				[]-> MayReturn =
						lists:append(element(5,H)--my_lookup(Key, element(5, H)), 	
									same_find(my_lookup(Key, element(5, H)), 
												element(2, element(1, H)))),
						[{element(1,H), element(2, H), element(3,H), element(4, H), MayReturn} | delete_key_cont(Key, T)];
				_-> MayReturn =
						lists:append(my_lookup(Key, element (2, H)),
							lists:append(element(5,H)--my_lookup(Key, element(5, H)), 	
									same_find(my_lookup(Key, element(5, H)), 
												element(2, element(1, H))))),
						[{element(1,H), element(2, H)--my_lookup(Key, element(2, H)), 
						  element(3,H), element(4, H), MayReturn}	
						| delete_key_cont(Key, T)]
			end
	end;
delete_key_cont(_Key, []) ->
	[].

delete_object_cont(Object,[H|T])->
	case lists:member(Object, element(2, H)++element(5, H)) of
		true ->	
			case lists:member(Object, element(2, H)) of
				true -> [{element(1, H), element(2, H)--[Object], 
						  element(3, H), element(4, H), [Object|element(5, H)]} 
						| delete_object_cont(Object, T)];
				false -> 
					case lists:member(Object, element(2, element(1, H))) of
						true-> [H | delete_object_cont(Object, T)];
						false-> [{element(1, H), element(2, H), element(3, H), element(4, H),
								 [Object|element(5, H)]} 
								| delete_object_cont(Object, T)]
					end
			end;
			
		false -> [ H | delete_object_cont(Object, T)]
	end;	
delete_object_cont(_Object, [])->
	[].
	
delete_pattern_cont(Pattern, [H|T])->
	case Pattern == element(3, H) of 
		true -> [{element(1, H),[], element(3, H), element(4, H), []} 
				|delete_pattern_cont(Pattern, T)];
		false -> [H|delete_pattern_cont(Pattern, T)]
	end;
delete_pattern_cont(_Pattern, []) ->
	[].
	
%%-------------------------------------------------------------------------------------------
%%------HERE THE LIST [H|T] REFERS TO THE matchcont IN THE state.
%%-------------------------------------------------------------------------------------------

delete_key_mcont(Key, [H|T]) ->
	case my_lookup(Key, element(2, H)++element(5, H))==[] of
		true -> [H | delete_key_mcont(Key, T)];
		false-> 
			case my_lookup(Key, element(2, H)) of
				[]-> [H | delete_key_mcont(Key, T)];
				_->	[{element(1,H), element(2, H)--my_lookup(Key, element(2, H)),
						element(3,H), element(4, H), 
						lists:append(element(5,H), my_lookup(Key, element(2, H)))} 
						| delete_key_mcont(Key, T)]
			end
	end;
delete_key_mcont(_Key, []) ->
	[].

delete_object_mcont(Object,[H|T])->
	case lists:member(Object, element(2, H)) of
		true -> Delete = element(2, H)--my_deleteobject(Object, element(2, H)),
				  [{element(1, H), my_deleteobject(Object, element(2, H)), element(3, H),	element(4, H), element(5, H)++Delete}
				    | delete_object_mcont(Object, T)];
		false -> [H	| delete_object_mcont(Object, T)]
	end;
delete_object_mcont(_Object, [])->
	[].

delete_pattern_mcont(Pattern, [H|T])->
	case Pattern == element(3, H) of 
		true -> [{element(1, H),[], element(3, H), element(4, H), 
					element(5, H)++element(2, H)}
					|delete_pattern_mcont(Pattern, T)];
		false -> [H|delete_pattern_mcont(Pattern, T)]
	end;
delete_pattern_mcont(_Pattern, []) ->
	[].
%%-------------------------------------------------------------------------------------------
%%-------------------------------------------------------------------------------------------
%%	THE FUNCTIONS BELOW ONLY FOR ETS TABLE.
%%-------------------------------------------------------------------------------------------
%%-------------------------------------------------------------------------------------------

my_tab2list(Tab)->
	ets:safe_fixtable(Tab, true),
		AllObjects=ets:tab2list(Tab),
	ets:safe_fixtable(Tab, false),
		AllObjects.
		
start_match_object(Tab, Pattern)->
	MCont=list_to_atom(io_lib:write(make_ref())),
	ReturnList=ets:match_object(Tab, Pattern),
	{MCont, ReturnList}.

finish_match_object({_MCont, ReturnList})->
	ReturnList.
	