-module(supervisor_tests).

-include_lib("eqc/include/eqc.hrl").
-include_lib("eqc/include/eqc_statem.hrl").
-include("pulse.hrl").

-compile(export_all).

-define(SUPERVISOR, supervisor).

-record(state, {relation, unrelation, supervisor, children, latestSupervisor}).	

%%-------------------------------------------------------------------------------------------
%%	relation: 
%%		is used to store the pairs of Id of supervisor and their children, which are called 
%%		by start_child() or restart_child(). EX: {Sup, Id, Pid}
%%	unrelation: 
%%		is used to store the pairs of Id of supervisor and their children which are 
%%		terminated. EX: {Sup, Id}
%%	supervisor: 
%%		is used to store the Id of the alive process with the type supervisor. 
%%	children:
%%		is used to store all the ChildSpec generated.
%%	latestSupervisor:
%%		is used to store the last modified supervisor.
%%-------------------------------------------------------------------------------------------

%%-------------------------------------------------------------------------------------------
%%	Generators
%%-------------------------------------------------------------------------------------------

gen_childSpec() ->					
	?LET(Id, gen_id(),
		?LET(Type, gen_type(),
		{Id, 
			case Type==?SUPERVISOR of
				true -> {?SUPERVISOR, start_link,[{local, Id}, supervisor_process, 
							gen_restart_strategy()]};
				false -> {supervisor_process, start, [Id]}
			end,
				gen_restart(),
					case Type==?SUPERVISOR of
						true-> infinity;
						false->	gen_shutDown()
					end,
						Type,
							[supervisor_process]})).

%%--------------------------	
gen_id() ->
	?LAZY(list_to_atom(io_lib:write(make_ref()))).

gen_restart() ->
	oneof([permanent, temporary, transient]).  

gen_shutDown() ->
	oneof([brutal_kill, 
			?SUCHTHAT(N, int(), N>0)]).

gen_type() ->
	oneof([?SUPERVISOR, worker]).
	
gen_restart_strategy()->
	oneof([one_for_one,
			one_for_all, 
				rest_for_one 
					]).
					
%%--------------------------

gen_latestSupervisor(S) ->
	frequency(
		[{2, begin
				[LatestSupervisor] = S#state.latestSupervisor,
				LatestSupervisor
			end} || S#state.latestSupervisor/=[]]++
		[{1, oneof(S#state.supervisor)}]
	).
	
%%-------------------------------------------------------------------------------------------
%%-------------------------------------------------------------------------------------------
%%
%%	Test function:	
%%		start_child(SupRef, ChildSpec) -> Result.
%%		terminate_child(SupRef, Id) -> Result.
%%		restart_child(SupRef, Id) -> Result.
%%		delete_child(SupRef, Id) -> Result.
%%		which_children(SupRef) -> All the children supervised by SupRef.
%%
%%-------------------------------------------------------------------------------------------
%%-------------------------------------------------------------------------------------------
%%	precondition.
%%-------------------------------------------------------------------------------------------
%%-------------------------------------------------------------------------------------------

precondition(S, {call, ?SUPERVISOR, start_child, [SupRef, _ChildSpec]}) ->		
	lists:member(SupRef, S#state.supervisor) == true;
precondition(S, {call, ?SUPERVISOR, terminate_child, [SupRef, Id]}) ->
	case lists:keyfind(Id, 2, S#state.relation) of
		{SupRef, Id, _, _}-> true;
		_-> false
	end;
precondition(S, {call, ?SUPERVISOR, delete_child, [SupRef, Id]}) ->	
	case lists:keyfind(Id, 2, S#state.unrelation) of
		{SupRef, Id, _, _}-> true;
		_-> false
	end;
precondition(S, {call, ?SUPERVISOR, restart_child, [SupRef, Id]}) ->
	case lists:keyfind(Id, 2, S#state.unrelation) of
		{SupRef, Id, _, _}-> true;
		_-> false
	end;
precondition(S, {call, ?SUPERVISOR, which_children, [SupRef]}) ->
	lists:member(SupRef, S#state.supervisor) == true;
precondition(S, {call, ?MODULE, erlang_exit, [Id, _Reason]}) ->
	lists:keymember(Id, 2, S#state.relation);
precondition(S, {call, ?MODULE, my_exit, [Id, _Reason]}) ->
	lists:keymember(Id, 2, S#state.relation).

%%-------------------------------------------------------------------------------------------
%%	next_state
%%-------------------------------------------------------------------------------------------

next_state(S, R, {call, ?SUPERVISOR, start_child, [SupRef, ChildSpec]}) ->		
	Pid={call, erlang, element, [2, R]},
	Id=element(1, ChildSpec),
	S1=S#state{children=[ChildSpec|S#state.children],
				relation=[{SupRef, Id, Pid, []}|S#state.relation],
				latestSupervisor=[SupRef]},
	case element(5, ChildSpec) of
		?SUPERVISOR -> S1#state{supervisor=[Id| S1#state.supervisor]};
		worker -> S1
	end;
next_state(S, _, {call, ?SUPERVISOR, terminate_child, [SupRef, Id]}) ->
	{L1, L2} = my_terminate([Id], S#state.relation, []),
	{SupRef, Id, Pid, OldPids}=lists:keyfind(Id, 2, S#state.relation),
	S#state{relation = L1--[lists:keyfind(Id, 2, S#state.relation)],
	         unrelation = S#state.unrelation ++ [{SupRef, Id, undefined, [Pid|OldPids]}],
			  supervisor = S#state.supervisor -- [Id|L2],
			   latestSupervisor = [SupRef]};
next_state(S, _, {call, ?SUPERVISOR, delete_child, [SupRef, Id]}) ->
	S#state{unrelation = S#state.unrelation--[lists:keyfind(Id, 2, S#state.unrelation)],
				latestSupervisor = [SupRef]};
next_state(S, R, {call, ?SUPERVISOR, restart_child, [SupRef, Id]}) ->
	Pid={call, erlang, element, [2, R]},
	{SupRef, Id, _, OldPids}=lists:keyfind(Id, 2, S#state.unrelation),
	S1=S#state{relation=S#state.relation ++ [{SupRef, Id, Pid, OldPids}],
			   unrelation=S#state.unrelation--[lists:keyfind(Id, 2, S#state.unrelation)]},
	case element(5, lists:keyfind(Id, 1, S1#state.children)) of
		?SUPERVISOR->S1#state{supervisor=S#state.supervisor ++[Id], 
					 latestSupervisor=[SupRef]};					
		worker->S1
	end;
next_state(S, R, {call, ?SUPERVISOR, which_children, [_Sup]})->
	S#state{relation=next_which_children(R, S#state.relation)};
next_state(S, _, {call, ?MODULE, erlang_exit, [Id, Reason]}) ->
	next_state_exit(S, Id, Reason);
next_state(S, _, {call, ?MODULE, my_exit, [Id, Reason]}) ->
	next_state_exit(S, Id, Reason).

%%-------------------------------------------------------------------------------------------
%%	postcondition
%%-------------------------------------------------------------------------------------------

postcondition(_S, {call, ?SUPERVISOR, start_child, [_SupRef, _ChildSpec]}, R) ->		
	element(1,R)==ok;	
postcondition(_S, {call, ?SUPERVISOR, terminate_child, [_SupRef, _Id]}, R) ->
	R==ok;
postcondition(_S, {call, ?SUPERVISOR, delete_child, [_SupRef, _Id]}, R) ->	
	R==ok;
postcondition(_S, {call, ?SUPERVISOR, restart_child, [_SupRef, _Id]}, R) ->
	element(1, R)==ok;
postcondition(S, {call, ?SUPERVISOR, which_children, [Sup]}, R)->
	post_which_children(R, S#state.relation, S#state.unrelation);
postcondition(S, {call, ?MODULE, erlang_exit, [Id, Reason]}, R)->
	true;
postcondition(S, {call, ?MODULE, my_exit, [Id, Reason]}, R)->
	true.
		
%%-------------------------------------------------------------------------------------------
%%	command 
%%-------------------------------------------------------------------------------------------

command(S) ->
	oneof([{call, ?SUPERVISOR, start_child, [oneof(S#state.supervisor), gen_childSpec()]}]++
		  [?LET({Sup, Id, _, _}, oneof(S#state.relation),
			{call, ?SUPERVISOR, terminate_child, [Sup, Id]}) || S#state.relation/=[]]++		
		  [?LET({Sup, Id, _, _}, oneof(S#state.unrelation),							
			{call, ?SUPERVISOR, delete_child, [Sup, Id]}) ||  S#state.unrelation/=[]]++	
		  [?LET({Sup, Id, _, _}, oneof(S#state.unrelation),
			{call, ?SUPERVISOR, restart_child, [Sup, Id]}) || S#state.unrelation/=[]]++
		  [?LET(LatestSupervisor, gen_latestSupervisor(S),
			{call, ?SUPERVISOR, which_children, [LatestSupervisor]}) ||S#state.latestSupervisor/=[]]++
		  [?LET({Sup, Id, _, _}, oneof(S#state.relation),
			case element(5, lists:keyfind(Id, 1, S#state.children))of
					?SUPERVISOR -> {call, ?MODULE, erlang_exit, [Id, kill]};
					worker-> {call, ?MODULE, my_exit, [Id, oneof([kill, normal])]}
			end)
			 || S#state.relation/=[]]
		  ).

%%-------------------------------------------------------------------------------------------
%%	property
%%-------------------------------------------------------------------------------------------

prop_supervisor_test() ->
	?FORALL(Sup, gen_id(), 
		begin
			?FORALL(Cmds, parallel2:pcommands(?MODULE, #state{relation = [], unrelation = [], 
							supervisor = [Sup], 
							children = [{Sup, {?SUPERVISOR, start_link, [{local, Sup}, supervisor_process, one_for_one]},temporary, 10000, ?SUPERVISOR, [supervisor_process]}], 
							latestSupervisor=[]}),	
				%?SCHEDULED([{verbose,false}], 
					begin
						%io:format("~p~n", [Cmds]),
						{ok, Pid}=?SUPERVISOR:start_link(supervisor_process, one_for_one),
						{ok, Pid2} = ?SUPERVISOR:start_child(Pid, 
									{Sup, {?SUPERVISOR, start_link, [{local, Sup}, supervisor_process, one_for_one]},temporary, 10000, ?SUPERVISOR, [supervisor_process]}),	
						X=parallel2:run_pcommands(?MODULE, Cmds, [{sup, Sup}]),
				%	end,
				%	X,
				%	begin
						kill_all_pids:kill_all_pids({Sup,X, Pid2}),
						{H,S,Res} = X,
						?WHENFAIL(begin
						    io:format("~p\n~p\n~p\n", [H, S, Res])
							%protest_pulse:draw_schedule(X)
						end,
							Res==ok)
					end)
		end).
		
%%-------------------------------------------------------------------------------------------
%%-------------------------------------------------------------------------------------------
%% Help function
%%-------------------------------------------------------------------------------------------
%%-------------------------------------------------------------------------------------------

my_terminate([H|T], Relation, List) ->
	case [X || {H1, X, _, _} <- Relation, H1==H] == [] of
		true -> my_terminate(T, Relation, List);
		false -> L1 = [{H1, X, Pid, OldPids} || {H1, X, Pid, OldPids} <- Relation, H1==H],
				 L2 = [X || {H1, X, _, _} <- Relation, H1==H],
				 my_terminate([ X || {H1, X, _, _} <- Relation, H1==H]++T, 			Relation--L1, List++L2)
	end;
my_terminate([], Relation, List) ->
	{Relation, List}.
%%-------The List here includes both the Id of supervisor and worker.
	
	

next_which_children([H|T], Relation)->
	case lists:keyfind(element(1, H), 2, Relation) of		
		{Sup, Id, unknown, OldPids}-> L=lists:keyreplace(element(1, H), 2,Relation,
												{Sup, Id, element(2, H), OldPids}),
										next_which_children(T, L);
		_->next_which_children(T, Relation)
	end;
next_which_children([], Relation)->
	Relation;
next_which_children({var, _}, Relation)->
	Relation.
	
	
	
	
post_which_children([H|T], Relation, Unrelation)->
%io:format("~n~npost_which_children: ~p ~p ~p ~p~n~n~n", [H,T,Relation,Unrelation]),
	case element(2, H) of
		undefined->
			case lists:keymember(element(1, H), 2, Unrelation) andalso
						lists:keymember(element(1, H), 2, Relation)==false of
				true-> post_which_children(T, Relation, Unrelation);
				false-> false
			end;
		_-> case (lists:member(element(2, H), 
						element(4, lists:keyfind(element(1, H), 2, Relation)))==false) andalso
				 (element(2, H)==eqc_symbolic:eval(element(3, lists:keyfind(element(1, H), 2, Relation))) orelse
				element(3, lists:keyfind(element(1, H), 2, Relation))==unknown )of
				true-> post_which_children(T, Relation, Unrelation);
				false-> false
			end
	end;
post_which_children([], Relation, Unrelaiton)->
	true.	
	
next_state_exit(S, Id, Reason)->
	{Sup, Id, _, _}=lists:keyfind(Id, 2, S#state.relation),
	{Sup, {_, _,[{_, _}, _,RestartStr]}, _, _, _, [_]}
								= lists:keyfind(Sup, 1, S#state.children),
	case Reason of 
		normal ->
			ChildSpecification=lists:keyfind(Id, 1, S#state.children),
			case element(3, ChildSpecification) of
				permanent -> 
					case RestartStr of
							one_for_one->
								S#state{relation=exit_restart([Id], S#state.relation)};
							one_for_all->
								L=[Id||{Sup, Id, _, _}<-S#state.relation],
								S#state{relation=exit_restart(L, S#state.relation)};
							rest_for_one->
								L=start_after(Id, S#state.relation, S#state.children, []),
								S#state{relation=exit_restart(L, S#state.relation)}
					end;  
				_ ->
					{L1, L2} = my_terminate([Id], S#state.relation, []),
					{SupRef, Id, Pid, OldPids}=lists:keyfind(Id, 2, S#state.relation),
					S#state{relation = L1--[lists:keyfind(Id, 2, S#state.relation)],
							unrelation = S#state.unrelation ++ [{SupRef, Id, undefined, [Pid|OldPids]}],
							supervisor = S#state.supervisor -- [Id|L2],
							latestSupervisor = [SupRef]}
			end;
		shutdown ->
			ChildSpecification=lists:keyfind(Id, 1, S#state.children),
			{L1, L2} = my_terminate([Id], S#state.relation, []),
			{SupRef, Id, Pid, OldPids}=lists:keyfind(Id, 2, S#state.relation),
			S#state{relation = L1--[lists:keyfind(Id, 2, S#state.relation)],
					unrelation = S#state.unrelation ++ [{SupRef, Id, undefined, [Pid|OldPids]}],
					supervisor = S#state.supervisor -- [Id|L2],
					latestSupervisor = [SupRef]};
		_->
			ChildSpecification=lists:keyfind(Id, 1, S#state.children),
			case element(3, ChildSpecification) of
				temporary -> 
					{L1, L2} = my_terminate([Id], S#state.relation, []),
					{SupRef, Id, Pid, OldPids}=lists:keyfind(Id, 2, S#state.relation),
					S#state{relation = L1--[lists:keyfind(Id, 2, S#state.relation)],
							unrelation = S#state.unrelation ++ [{SupRef, Id, undefined, [Pid|OldPids]}],
							supervisor = S#state.supervisor -- [Id|L2],
							latestSupervisor = [SupRef]};
				_ -> 
					case RestartStr of
							one_for_one->
								S#state{relation=exit_restart([Id], S#state.relation)};
							one_for_all->
								L=[Id||{Sup, Id, _, _}<-S#state.relation],
								S#state{relation=exit_restart(L, S#state.relation)};
							rest_for_one->
								L=L=start_after(Id, S#state.relation, S#state.children, []),
								S#state{relation=exit_restart(L, S#state.relation)}
					end
			end
	end.

	
exit_restart([H|T], Relation)->
	{Sup, H, Pid, OldPids}=lists:keyfind(H, 2, Relation),
	L=lists:keyreplace(H, 2, Relation,{Sup, H, unknown, [Pid|OldPids]}),
	exit_restart(T, L);
exit_restart([], Relation)->
	Relation.		

	
start_after(Id, Relation, [H|T], L)->
	{Sup, _, _, _}=lists:keyfind(Id, 2, Relation),
	case element(1, H)==Id of
		true -> start_after(Id, Relation, [], [Id|L]);
		false->
			case lists:member(element(1, H), my_whichChildren(Sup, Relation)) of 
				true ->start_after(Id, Relation, T, [element(1, H)|L]);
				false->start_after(Id, Relation, T, L)
			end
	end;
start_after(_Id, _Relation, [], L)->
	L.	
	
my_whichChildren(Sup, Relation)->
	[X || {Sup1, X} <- Relation, Sup==Sup1].
	
	
%my_exit(Id, Reason)->
%		Pid = whereis(Id),
%		Ref = monitor(process, Pid),
%		Id ! {exit, Reason},
%		receive
%			{'DOWN', Ref, process, _, _} ->
%				ok
%		end.
my_exit(Id, Reason) ->
	Pid = whereis(Id),
	case Pid of
	  undefined -> ok;
	  _ -> Ref = monitor(process, Pid),
		   Pid ! {exit, Reason},
	       receive {'DOWN', Ref, _, _, _} -> ok after 1000 -> exit(timed_out) end
	end.
	
erlang_exit(Id, Reason) ->
	Pid = whereis(Id),
	case Pid of
	  undefined -> ok;
	  _ -> Ref = monitor(process, Pid),
		   exit(Pid, Reason),
		   receive after 100 -> ok end,
	       receive {'DOWN', Ref, _, _, _} -> ok after 1000 -> exit(timed_out) end
	end.
	
compile()->
	instrument:c(genserver),
	instrument:c(supervisor_tests),
	instrument:c(super_visor),
	instrument:c(supervisor_process),
	instrument:c(parallel2),
	error_logger:tty(false).
	
f() ->
	 catch unregister(sup),
	 catch unregister(sup2),
	 {ok, Pid} = supervisor:start_link(supervisor_process, one_for_one),
	 supervisor:start_child(Pid, {sup, {supervisor, start_link, [{local, sup}, supervisor_process, one_for_one]}, temporary, 10000, supervisor, [supervisor_process]}),
	 supervisor:start_child(sup, {sup2, {supervisor, start_link, [{local, sup2}, supervisor_process, one_for_all]}, temporary, infinity, supervisor, [supervisor_process]}),
	 Sup = whereis(sup),
	 Sup2 = whereis(sup2),
	 %Ref = monitor(process, Sup2), 
	erlang:exit(Sup2, kill),
	%receive {'DOWN', Ref, _, _, _} -> ok end,
	 %receive after 1000-> ok end,
	 %{Sup, Sup2, process_info(Sup, messages)}.
	 supervisor:delete_child(sup, sup2).